#include <sys/time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>

#include "TimeFns.h"
#include "MNLDaemon.h"
#include "SendQueue.h"

#include "PeriodicPacketDropper.h"
#include "SpecificPacketDropper.h"
#include "ProbabilisticPacketDropper.h"
 

bool MNLDaemon::Init(const char* ipCfgFileName) {

  if (!ReadCfgFile(ipCfgFileName)) {
    return false;
  }

  if (!InitListenSkt()) {
    return false;
  }

  mpSendQueue = new SendQueue(mRTTMean, mRTTVar, mpPktDropper);
  if (!mpSendQueue->Init()) {
    return false;
  }

  return true;
}

bool MNLDaemon::InitListenSkt() {
  
  mListenSkt = socket(PF_INET, SOCK_DGRAM, 0);
  if (mListenSkt == -1) {
    printf ("MNL: Unable to open UDP listen socket: %s\n",
            strerror(errno)
          );
    return false;
  }

  int lYes = 1;
  if (setsockopt(
                  mListenSkt,
                  SOL_SOCKET, SO_REUSEADDR, 
                  &lYes, sizeof(int)
                ) == -1) {
    printf ("MNL: Unable to set SO_REUSEADDR option on udp listen socket: %s.\n",
              strerror (errno)
            );
    return false;
  }
 
  struct sockaddr_in lMyAddr;
  socklen_t lMyAddrLen = sizeof(lMyAddr);
  // Bind to a random port
  lMyAddr.sin_family = AF_INET; 
  lMyAddr.sin_port = 0;
  lMyAddr.sin_addr.s_addr = INADDR_ANY;
  memset(&(lMyAddr.sin_zero), '\0', sizeof(lMyAddr.sin_zero));

  if (bind( mListenSkt , 
            (struct sockaddr *)&lMyAddr, 
            lMyAddrLen
            ) == -1) {

    printf ( "MNL: Unable to bind udp listen socket to a random port: %s\n",
              strerror (errno)
            );
    return false;
  }

  // Get the port on which we are listening.
  if ( getsockname (
                    mListenSkt,
                    (struct sockaddr *) &lMyAddr, 
                    &lMyAddrLen
                    ) != -1) {

    printf("MNL: Listening on %s:%d\n",
                    inet_ntoa (lMyAddr.sin_addr),
                    ntohs (lMyAddr.sin_port)
                );

  } else {
    printf ("MNL: Error: Unable to get port number of listening socket :%s\n",
            strerror(errno)
          );
    return false;
  }

  //write the listening port number to the MNLDaemon sock file
  FILE* lFd = fopen(MNL_SOCK_FILE, "w");
  if (lFd == NULL) {
    printf ("MNL: Init(): Unable to open %s for writing: %s\n",
            MNL_SOCK_FILE,
            strerror(errno)
          );
    return false;
  }
  fprintf(lFd, "%d", ntohs(lMyAddr.sin_port));
  fclose(lFd);
   
  return true;

}

void MNLDaemon::MainLoop() {
  
  struct timeval lNextTimerEvent;
  lNextTimerEvent.tv_sec = 0;
  lNextTimerEvent.tv_usec = 0;

  struct timeval lNextTimerDiff;

  while (true) {

    struct timeval lTimeToSleep = mpSendQueue->GetTimeToNextSend();
    
    struct timeval* lpTimeToSleep = NULL;
    if (lTimeToSleep.tv_sec == 0 && lTimeToSleep.tv_usec == 0) {

      if (mpSendQueue->IsEmpty()) {
        lpTimeToSleep = NULL;
      } else {
        mpSendQueue->TransmitAll();
        continue;
      }

    } else {
      lpTimeToSleep = &lTimeToSleep;
    }
    
    fd_set lRdFds;
    FD_ZERO(&lRdFds);
    FD_SET(mListenSkt, &lRdFds);

    int lSelectRetVal = select (
                            mListenSkt + 1,
                            &lRdFds,
                            NULL,
                            NULL,
                            lpTimeToSleep
                        );
    if (lSelectRetVal < 0) {
      
      printf ("MNL: Daemon: Error in main loop: %s\n",
                strerror(errno)
            );
      exit (1);
      
    } else if (lSelectRetVal == 1 && FD_ISSET(mListenSkt, &lRdFds)) {
      
      ProcessClientMsg();
    
    } else if (lSelectRetVal == 0) {
      //Timer event. we will process this later,
    } else {
      printf ("MNL: Warning: Invalid return value for select in main loop : %d\n",
              lSelectRetVal
            );
    }

    mpSendQueue->TransmitAll();
  }
}

/**
  * @return 0 on error. 1 on success.
  */
bool MNLDaemon::ReadCfgFile(const char* ipCfgFileName) {

  char lpLine[10000];
  char* lpDelims = ":";
  char* lpToken = NULL;

  FILE* lFd = fopen(ipCfgFileName, "r");
  if (lFd == NULL) {
    printf ("MNL: Error: Unable to open configuration file %s: %s\n", ipCfgFileName, strerror(errno));
    return false;
  }

  if (fgets(lpLine, sizeof(lpLine), lFd) == NULL) {
    printf("MNL: Fatal Error: while reading RTTMean from the configuration file %s:%s\n", ipCfgFileName, strerror(errno));
    fclose(lFd);
    return false;
  }
  mRTTMean = atoi(lpLine);
  
  if (fgets(lpLine, sizeof(lpLine), lFd) == NULL) {
    printf("MNL: Fatal Error: while reading RTTVar from the configuration file %s:%s\n", ipCfgFileName, strerror(errno));
    fclose(lFd);
    return false;
  }
  mRTTVar = atoi(lpLine);
 
  if (fgets(lpLine, sizeof(lpLine), lFd) == NULL) {
    printf("MNL: Fatal Error: while reading drop information from the configuration file %s:%s\n", ipCfgFileName, strerror(errno));
    fclose(lFd);
    return false;
  }

  char* lpLineSave = strdup (lpLine);
  lpToken = strtok( lpLine, lpDelims );
  if (lpToken == NULL) {
    printf ("MNL: Error: Invalid line : %s\n", lpLine);
    fclose(lFd);
    return false;
  }

  if (strcmp(lpToken, "drop_none") == 0) {
    printf ("MNL: DROP_NONE\n");
    mpPktDropper = NULL;

  } else if (strcmp(lpToken, "drop_specific") == 0) {
    
    printf ("MNL: DROP_SPECIFIC\n");
    mpPktDropper = new SpecificPacketDropper(lpLineSave);
    
  } else if (strcmp(lpToken, "drop_periodic") == 0) {

    printf ("MNL: DROP_PERIODIC\n");
    mpPktDropper = new PeriodicPacketDropper(lpLineSave);
    
  } else if (strcmp(lpToken, "drop_probabilistic") == 0) {
    
    printf ("MNL: DROP_PROBABILISTIC\n");
    mpPktDropper = new ProbabilisticPacketDropper(lpLineSave);

  } else {
    printf ("MNL: Error: Invalid drop type : %s\n", lpLine);
    fclose(lFd);
    return false;
  }

  free(lpLineSave);
  fclose(lFd);

  return true;

}

void MNLDaemon::ProcessClientMsg() {

  //printf ("MNL: ProcessClientMsg: Recd\n");

  static uint8_t* lpRecvBuf = NULL;
  static int lRecvBufLen = 0;
  
  if (lpRecvBuf == NULL) {
    lRecvBufLen = sizeof(SendMsg) + MAX_PKT_LEN;
    lpRecvBuf = (uint8_t*) malloc (lRecvBufLen * sizeof(uint8_t));
    assert (lpRecvBuf != NULL);
  }
  struct sockaddr_in lRemoteAddr;
  socklen_t lRemoteAddrLen = sizeof(struct sockaddr_in);
  int lNumBytesRead = recvfrom (  
                                  mListenSkt, 
                                  lpRecvBuf, 
                                  lRecvBufLen,
                                  0, 
                                  (struct sockaddr *)&lRemoteAddr, 
                                  &lRemoteAddrLen
                              );
  if (lNumBytesRead == -1) {
    printf ("MNL: ProcessClientMsg: Error while receiving UDP packet : %s.\n", strerror (errno));
    return;
  }
       
  if (lNumBytesRead > 0) {
    mpSendQueue->AddPacket((SendMsg *) lpRecvBuf);
  } else {
    printf ("MNL: ProcessClientMsg: Warning: 0 byte received from %s:%d\n",
            inet_ntoa(lRemoteAddr.sin_addr),
            ntohs(lRemoteAddr.sin_port)
        );
  }

}

MNLDaemon::~MNLDaemon() {
  
  unlink (MNL_SOCK_FILE);

}
