#include <sys/time.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>


#include "TimeFns.h"
#include "SendQueue.h"
#include "PacketDropper.h"

#include "MNLClient.h"

/**
  * Initializes the UDP socket over which the packets are finally
  * transmitted.
  */
bool SendQueue::Init() {


  mTxSkt = socket(PF_INET, SOCK_DGRAM, 0);
  if (mTxSkt == -1) {
    printf ("MNL: SendQueue: Init(): Unable to open UDP transmit socket: %s\n",
            strerror(errno)
          );
    return false;
  }

  int lYes = 1;
  if (setsockopt(
                  mTxSkt,
                  SOL_SOCKET, SO_REUSEADDR, 
                  &lYes, sizeof(int)
                ) == -1) {
    printf ("MNL: Unable to set SO_REUSEADDR option on udp transmit socket: %s.\n",
              strerror (errno)
            );
    return false;
  }
 
  struct sockaddr_in lMyAddr;
  // Bind to a random port
  lMyAddr.sin_family = AF_INET; 
  lMyAddr.sin_port = 0;
  lMyAddr.sin_addr.s_addr = INADDR_ANY;
  memset(&(lMyAddr.sin_zero), '\0', sizeof(lMyAddr.sin_zero));

  if (bind( mTxSkt , 
            (struct sockaddr *)&lMyAddr, 
            sizeof(struct sockaddr_in)
            ) == -1) {

    printf ( "MNL: SendQueue: Init(): Unable to bind udp transmit socket to a random port: %s\n",
              strerror (errno)
            );
    return false;
  }

  return true;

}

void SendQueue::AddPacket(SendMsg* ipSendMsg) {
 
  struct timeval lCurrTime;
  gettimeofday(&lCurrTime, NULL);

  float lRand = rand() * 1.0/RAND_MAX;
  
  float lCurrRTT = mRTTMean /2.0 + ((0.5 - lRand) * mRTTVar)/2;
  int lCurrRTTInt = (int) lCurrRTT;

  if (lCurrRTTInt <= 0) {
    printf ("MNL: Invalid specification of RTT and RTTVar\n");
    exit (1);
  }

  struct timeval lDelayTime;
  lDelayTime.tv_sec = lCurrRTTInt /1000000;
  lDelayTime.tv_usec = lCurrRTTInt % 1000000;
  
  struct timeval lToSendTime = TimeAdd(lCurrTime, lDelayTime);
      

  PacketBuf* lpPktBuf = new PacketBuf(ipSendMsg, lToSendTime);

  mPktQueue.push(lpPktBuf);
}

bool SendQueue::IsEmpty() {
  return mPktQueue.empty();
}

/**
  * Transmit the specified packet.
  */
void SendQueue::Transmit(PacketBuf* ipPkt) {

  printf ("MNL: SendQueue: Transmit: Transmitting pkt (%d bytes, seqNum=%d, len=%d) scheduled for time %d:%d to %s:%d\n", 
          ipPkt->mPktLen,
          ipPkt->mSeqNumStart,
          ipPkt->mDataLen,
          ipPkt->mToSendTime.tv_sec,
          ipPkt->mToSendTime.tv_usec,
          inet_ntoa(ipPkt->mToAddr.sin_addr),
          ntohs(ipPkt->mToAddr.sin_port)
        );

  bool lDropFlag = false;
  
  if (mpPktDropper != NULL) {
    lDropFlag = mpPktDropper->IsDropPkt();

    // This refers to the packet as sent by the app.
    // Note that we are counting dropped packets also here.
    mpPktDropper->PktSent();
 
  }
  
  if (lDropFlag) {
    printf ("MNL: Dropping packet\n");
    return;
  }

  // Do the actual UDP send
  int lRetVal = sendto( mTxSkt, 
                    ipPkt->mpPktContents, 
                    ipPkt->mPktLen, 
                    ipPkt->mFlags, 
                    (struct sockaddr *) &(ipPkt->mToAddr),
                    ipPkt->mToLen
                  );

  if (lRetVal < 0) {
    
    printf ("MNL: SendQueue:Transmit(): Error while sending packet to %s:%d - %s\n",
              inet_ntoa(ipPkt->mToAddr.sin_addr),
              ntohs(ipPkt->mToAddr.sin_port),
              strerror(errno)
            );
    return;
    
  } else if (lRetVal != ipPkt->mPktLen) {
    
    printf ("MNL: SendQueue:Transmit(): Warning: Did not send the whole packet (only %d of %d sent).\n",
        lRetVal,
        ipPkt->mToLen
        );
    return;
  }
}

/** 
  * Transmit and delete all packets whose ToSendTime is less than the current time.
  */
void SendQueue::TransmitAll() {

  struct timeval lCurrTime;
  gettimeofday(&lCurrTime, NULL);

  while ( ! mPktQueue.empty()) {
    
    PacketBuf* lpCurrPkt = mPktQueue.top();
    if (TimeCmp(lpCurrPkt->mToSendTime, lCurrTime) <= 0) {
      Transmit(lpCurrPkt);
      mPktQueue.pop();
      delete(lpCurrPkt);
    } else {
      // Remember this is a priority queue
      break;
    }
  }
}

struct timeval SendQueue::GetTimeToNextSend() {

  struct timeval lRetVal;
  lRetVal.tv_sec = 0;
  lRetVal.tv_usec = 0;
  
  if (mPktQueue.empty()) {
    return lRetVal;
  }

  struct timeval lCurrTime;
  gettimeofday(&lCurrTime, NULL);

  PacketBuf* lpPkt = mPktQueue.top();
  lRetVal = TimeDiff(lpPkt->mToSendTime, lCurrTime);
  
  return lRetVal;
}

SendQueue::SendQueue(uint32_t iRTTMean, uint32_t iRTTVar, PacketDropper* ipPktDropper) {

  mRTTMean = iRTTMean;
  mRTTVar = iRTTVar;
  mpPktDropper = ipPktDropper;

}
