#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#include <string.h>
#include <errno.h>

#include <sys/socket.h>
#include <fcntl.h>

#include <arpa/inet.h>
#include <netinet/in.h>

// Header file required for uint8_t, uint16_t etc.
#include <sys/types.h>


/**
  * Prints the usage information.
  */
void usage (const char* ipProgName) {
    printf (
                "\nUsage: %s <listen port>\n"
                "\n"
                ,
                ipProgName
            );
}

int gUDPListenSkt = 0;
int gUDPListenPort = 0;

/**
  * Creates a UDP socket and binds to the udp listen port
  * @return The UDP socket fd.
  * Aborts the program if some error occurs.
  */
int InitUDP () {
    
    int tmp_skt = socket(PF_INET, SOCK_DGRAM, 0);
    if (tmp_skt == -1) {
        printf (
                        "Unable to open udp listen socket: %s.\n",
                        strerror (errno)
                    );
        exit (1);
    }

    int yes=1;
    if (setsockopt(tmp_skt, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
        printf (
                        "Unable to set SO_REUSEADDR option on udp listen socket: %s.\n",
                        strerror (errno)
                    );
        exit (1);
    }
    
    struct sockaddr_in my_addr;    // my address information
    my_addr.sin_family = AF_INET;    // host byte order 
    my_addr.sin_port = htons (gUDPListenPort) ;     // short, network byte order
    my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
    memset(&(my_addr.sin_zero), '\0', sizeof(my_addr.sin_zero)); // zero the rest of the struct

    if (bind( tmp_skt , (struct sockaddr *)&my_addr, sizeof(struct sockaddr_in)) == -1) {
        printf (
                        "Unable to bind udp listen socket to port %d : %s.\n",
                        gUDPListenPort, strerror (errno)
                    );
        exit (1);
    }

    return tmp_skt;
}

void SendReply(uint16_t lReplyPortN, char* ipPkt, int iPktLen) {
  
  struct sockaddr_in dst_addr;
  dst_addr.sin_family = AF_INET;     // host byte order
  dst_addr.sin_port = lReplyPortN;
  dst_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  memset(&(dst_addr.sin_zero), '\0', sizeof(dst_addr.sin_zero));  // zero the rest of the struct


  int numSent = sendto (gUDPListenSkt, ipPkt, iPktLen, 0,
                                  (struct sockaddr *) &dst_addr, sizeof(struct sockaddr_in)
                              );
  if (numSent == -1) {
    printf ("Error while sending reply: %s\n",
            strerror(errno)
          );
  } else if (numSent < iPktLen) {
    printf ("Error: Couldn't set the whole reply\n");
  }

}

void RecvUDPPkt() {

    char recvBuf [10000];
    int recvBufLen = sizeof(recvBuf);
   
    struct sockaddr_in remoteAddr;
    socklen_t addrLen = sizeof(struct sockaddr_in);
    int numBytesRead = recvfrom (gUDPListenSkt, recvBuf, recvBufLen,0, (struct sockaddr *)&remoteAddr, &addrLen);
    if (numBytesRead == -1) {
        printf ("Error while receiving UDP packet : %s.\n", strerror (errno));
        return;
    }
       
    if (numBytesRead > 0) {
        // Some bytes were actually read.
        // Echo it back
       
        uint16_t lResponsePortN;
        memcpy (&lResponsePortN, recvBuf, sizeof(lResponsePortN));

        // sanitize the received string in order to prevent buffer overflow if
        // dangerous string functions like strcpy are used.
        recvBuf [recvBufLen - 1] = '\0';
        printf ("%s:%d Received %d bytes over UDP: %s\n", 
                inet_ntoa (remoteAddr.sin_addr),
                ntohs(remoteAddr.sin_port),
                numBytesRead,
                recvBuf + 2
              );

        SendReply (lResponsePortN, recvBuf + 2, numBytesRead);

    } else {
        printf ("%s:%d Warning: 0 bytes received.\n", 
                            inet_ntoa (remoteAddr.sin_addr),
                            ntohs (remoteAddr.sin_port)
                    );
    }

}

/** Called when the program is stopped. */

void DoCleanup (int signum) { 

  printf ("Closing the UDP Listen Socket %d.\n", gUDPListenSkt);
  close (gUDPListenSkt);
  
  exit (0);
}

/**
  * The program begins here.
  */
int main (int argc, char* argv[]) {

    // Catch Ctrl-C
    signal (SIGINT, DoCleanup);

    if (argc < 2) {
      usage(argv[0]);
      exit(1);
    }
    
    gUDPListenPort = atoi (argv[1]);
    

    printf (
                "\n"
                "Starting TestRecv UDP Port %d...\n",
                gUDPListenPort
            );

    gUDPListenSkt = InitUDP();

    // Listen indefinitely
    while (1) {
      RecvUDPPkt();
    }
}

